/*
** OS-Specific code for Solaris 2.x for oidentd
** ultr0s <ultros@ojnk.net>
**
** Includes code blatantly stolen from pidentd's svr4.c and sunos5.c
**
** $Id: solaris2.c,v 1.8 2001/07/25 17:15:31 odin Exp $
*/

#include <config.h>

#define _KMEMUSER
#define _KERNEL

#define exit		kernel_exit
#define strsignal	kernel_strsignal

#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/param.h>
#include <netinet/in.h>

#include <stdio.h>
#include <kvm.h>
#include <nlist.h>
#include <math.h>
#include <syslog.h>
#include <stddef.h>
#include <sys/fcntl.h>
#include <sys/cred.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <inet/common.h>
#include <inet/ip.h>

#include <oidentd.h>
#include <oidentd_util.h>
#include <oidentd_inet_util.h>

#define N_FANOUT 0

#if SOLARIS >=6
#	define FANOUT_OFFSET(n)	(nl[N_FANOUT].n_value + \
				(n) * sizeof(icf_t) + offsetof(icf_t, icf_ipc))

#	define SOL_TCP_FO_STR	"ipc_tcp_conn_fanout"
#else
#	define FANOUT_OFFSET(n)	(nl[N_FANOUT].n_value + (n) * sizeof(ipc_t *))
#	define SOL_TCP_FO_STR	"ipc_tcp_fanout"
#endif

#undef exit
#undef strsignal

static int getbuf(u_long addr, void *kbuf, size_t len);

static struct nlist nl[] = {
	{ SOL_TCP_FO_STR },
	{ 0 },
};

static kvm_t *kd;

int get_user4(	in_port_t lport,
				in_port_t fport,
				struct sockaddr_storage *laddr,
				struct sockaddr_storage *faddr)
{
	queue_t sqr;
	ipc_t ic, *icp;
	in_port_t uslp, usfp;
	u_int32_t offset;
	file_t tf;
	u_long zero = 0;
	in_port_t *ports;
	in_addr_t *locaddr, *raddr;
	struct proc *procp;

	usfp = fport;
	uslp = lport;

	offset = usfp ^ uslp;
	offset ^= (u_int32_t) SIN4(faddr)->sin_addr.S_un.S_un_b.s_b4 ^ (offset >> 8);
	offset &= 0xff;

	if (getbuf(FANOUT_OFFSET(offset), &icp, sizeof(ipc_t *)) == -1)
		return (-1);

#if SOLARIS >= 6
	locaddr = (in_addr_t *) &ic.ipc_laddr;
	raddr = (in_addr_t *) &ic.ipc_faddr;
	ports = (in_port_t *) &ic.ipc_ports;
#else
	locaddr = &ic.ipc_tcp_laddr;
	raddr = &ic.ipc_tcp_faddr;
	ports = (in_port_t *) &ic.ipc_tcp_ports;
#endif

	while (icp != NULL) {
		if (getbuf((u_long) icp, &ic, sizeof(ic)) == -1)
			return (-1);

		if (usfp == ports[0] && uslp == ports[1] &&
			(SIN4(laddr)->sin_addr.s_addr == *locaddr ||
			SIN4(faddr)->sin_addr.s_addr == *raddr) &&
			SIN4(faddr)->sin_addr.s_addr == *raddr)
		{
			break;
		}

		icp = ic.ipc_hash_next;
	}

	if (icp == NULL)
		return (-1);

	if (getbuf((u_long) ic.ipc_rq + offsetof(queue_t, q_stream),
				&sqr.q_stream, sizeof(sqr.q_stream)) == -1)
		return (-1);

	if (kvm_setproc(kd) != 0)
		return (-1);

	while ((procp = kvm_nextproc(kd)) != NULL) {
		struct uf_entry files[NFPCHUNK];
		int nfiles = procp->p_user.u_nofiles;
		u_long addr = (u_long) procp->p_user.u_flist;

		while (nfiles > 0) {
			int nread = nfiles > NFPCHUNK ? NFPCHUNK : nfiles;
			size_t size = nread * sizeof(struct uf_entry);
			int i;
			int ret;
			struct file *last = NULL;
			vnode_t vp;

			if (getbuf(addr, &files[0], size) == -1)
				return (-1);

			for (i = 0 ; i < nread ; i++) {
				if (files[i].uf_ofile == 0 || files[i].uf_ofile == last)
					continue;

				last = files[i].uf_ofile;
				if (getbuf((u_long) last, &tf, sizeof(tf)) == -1)
					return (-1);

				if (!tf.f_vnode)
					continue;

				ret = getbuf((u_long) tf.f_vnode + offsetof(vnode_t, v_stream),
							&vp.v_stream,
							sizeof(vp.v_stream));

				if (ret == -1)
					return (-1);

				if (vp.v_stream == sqr.q_stream) {
					cred_t cr;

					ret = getbuf((u_long) tf.f_cred + offsetof(cred_t, cr_ruid),
					 			&cr.cr_ruid,
					 			sizeof(cr.cr_ruid));

					if (ret == -1)
						return (-1);

					return (cr.cr_ruid);
				}
			}

			nfiles -= nread;
			addr += size;
		}
	}

	return (-1);
}

static int getbuf(u_long addr, void *kbuf, size_t len) {
	int i;

	for (i = 0 ; i < 5 ; i++) {
		if (kvm_read(kd, addr, kbuf, len) >= 0)
			return (0);
	}

	return (-1);
}

int k_open(void) {
	kd = kvm_open(NULL, "/dev/mem", NULL, O_RDONLY, NULL);

	if (kd == NULL) {
		o_log(DEBUG, "Error: kvm_open: %s", strerror(errno));
		return (-1);
	}

	if (kvm_nlist(kd, nl) != 0) {
		o_log(DEBUG, "Error: kvm_nlist: %s", strerror(errno));
		return (-1);
	}

	return (0);
}
